#include <algorithm>
#include <cstring>
#include <set>
#include <msp/gl/extensions/arb_uniform_buffer_object.h>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include "extension.h"
#include "program.h"
#include "programbuilder.h"
#include "shader.h"
#include "vertexformat.h"

using namespace std;

namespace Msp {
namespace GL {

/*
Naming conventions:
  n_*        Normalized vector
  l_*        Lighting component

  obj_*      Object space
  eye_*      Eye space
  tbn_*      Tangent-Binormal-Normal space
  shd_*      Shadow space
  env_*      Environment space
  *_dir      Direction vector

  zzz_*      Wildcard space, resolved by the builder
             All wildcard spaces within an expression must match

  xxx_yyy_*  Matrix that transforms between yyy to xxx
             The vector is on the side of its designated space, result will be
             in the other space
  *_matrix   A matrix (duh)
  *_rmatrix  A mat4 that works with a row vector

  rgb_*      Color with rgb components only
  color_*    Color with rgba components
*/

/* The array is stored in reverse order, so that variables always come after
anything that might need them. */
const ProgramBuilder::VariableDefinition ProgramBuilder::standard_variables[] =
{
	{ FRAGMENT, "gl_FragColor", "vec4", "frag_color", 0 },
	{ FRAGMENT, "frag_color", "vec4", "incoming_color", "!f" },
	{ FRAGMENT, "frag_color", "vec4", "vec4(mix(fog_color.rgb, incoming_color.rgb, fog_value), incoming_color.a)", "f" },
	{ FRAGMENT, "fog_value", "float", "exp(-fog_coord*fog_density)", 0 },

	{ FRAGMENT, "incoming_color", "vec4", "basic_color", "!e!l" },
	{ FRAGMENT, "incoming_color", "vec4", "vec4(rgb_surface, surface_alpha)", "!el" },
	{ FRAGMENT, "incoming_color", "vec4", "vec4(rgb_surface+rgb_reflection, surface_alpha)", "e" },

	{ FRAGMENT, "rgb_reflection", "vec3", "reflect_sample.rgb*reflectivity", 0 },
	{ FRAGMENT, "reflect_sample", "vec4", "textureCube(environment, env_reflect_dir)", 0 },
	{ FRAGMENT, "env_reflect_dir", "vec3", "env_eye_matrix*eye_reflect_dir", 0 },
	{ FRAGMENT, "eye_reflect_dir", "vec3", "eye_tbn_matrix*tbn_reflect_dir", "n" },

	{ FRAGMENT, "rgb_surface", "vec3", "rgb_unlit_surface", "!l" },
	{ FRAGMENT, "rgb_surface", "vec3", "rgb_lit_surface", "l" },
	{ FRAGMENT, "rgb_unlit_surface", "vec3", "basic_color.rgb", "!s" },
	{ FRAGMENT, "rgb_unlit_surface", "vec3", "basic_color.rgb*l_shadow", "s" },
	{ FRAGMENT, "rgb_lit_surface", "vec3", "rgb_illumination_diffuse", "!m!t" },
	{ FRAGMENT, "rgb_lit_surface", "vec3", "rgb_illumination_diffuse*diffuse_sample.rgb", "!mt" },
	{ FRAGMENT, "rgb_lit_surface", "vec3", "rgb_surface_ambient+rgb_surface_diffuse", "m!p" },
	{ FRAGMENT, "rgb_lit_surface", "vec3", "rgb_surface_ambient+rgb_surface_diffuse+rgb_surface_specular", "mp" },

	{ FRAGMENT, "surface_alpha", "float", "basic_color.a", "!m" },
	{ FRAGMENT, "surface_alpha", "float", "basic_color.a", "m!l" },
	{ FRAGMENT, "surface_alpha", "float", "material.diffuse.a", "ml!t" },
	{ FRAGMENT, "surface_alpha", "float", "material.diffuse.a*diffuse_sample.a", "mlt" },
	{ FRAGMENT, "basic_color", "vec4", "vec4(1.0)", "!m!t" },
	{ FRAGMENT, "basic_color", "vec4", "diffuse_sample", "!mt" },
	{ FRAGMENT, "basic_color", "vec4", "color", "m!t" },
	{ FRAGMENT, "basic_color", "vec4", "color*diffuse_sample", "mt" },

	{ FRAGMENT, "rgb_surface_ambient", "vec3", "rgb_illumination_ambient*material.ambient.rgb", "!t" },
	{ FRAGMENT, "rgb_surface_ambient", "vec3", "rgb_illumination_ambient*material.ambient.rgb*diffuse_sample.rgb", "t" },
	{ FRAGMENT, "rgb_illumination_ambient", "vec3", "ambient_color.rgb", "!y" },
	{ FRAGMENT, "rgb_illumination_ambient", "vec3", "ambient_color.rgb+l_skylight*sky_color.rgb", "y" },
	{ FRAGMENT, "l_skylight", "float", "dot(n_zzz_normal, zzz_sky_dir)*0.5+0.5", 0 },

	{ FRAGMENT, "rgb_surface_specular", "vec3", "rgb_illumination_specular*material.specular.rgb", 0 },
	{ FRAGMENT, "rgb_illumination_specular", "vec3", "rgb_illumination_specular_direct", "!y|e" },
	{ FRAGMENT, "rgb_illumination_specular", "vec3", "rgb_illumination_specular_direct+l_sky_specular*sky_color.rgb", "y!e" },
	{ FRAGMENT, "rgb_illumination_specular_direct", "vec3", "rgb_light_specular", "!s" },
	{ FRAGMENT, "rgb_illumination_specular_direct", "vec3", "rgb_light_specular*l_shadow", "s" },
	{ FRAGMENT, "rgb_light_specular[i]", "vec3", "l_specular[i]*light_sources[i].specular.rgb", 0 },
	{ FRAGMENT, "l_sky_specular", "float", "pow((1.0-pow(clamp(dot(zzz_reflect_dir, zzz_sky_dir)-horizon_limit, -1.0, 0.0), 2.0)), material.shininess/2.0)", 0 },
	{ FRAGMENT, "l_specular[i]", "float", "pow(max(dot(n_zzz_half_vec[i], n_zzz_normal), 0.0), material.shininess)", 0 },

	{ FRAGMENT, "rgb_surface_diffuse", "vec3", "rgb_illumination_diffuse*material.diffuse.rgb", "!t" },
	{ FRAGMENT, "rgb_surface_diffuse", "vec3", "rgb_illumination_diffuse*material.diffuse.rgb*diffuse_sample.rgb", "t" },
	{ FRAGMENT, "rgb_illumination_diffuse", "vec3", "rgb_light_diffuse", "!s" },
	{ FRAGMENT, "rgb_illumination_diffuse", "vec3", "rgb_light_diffuse*l_shadow", "s" },
	{ FRAGMENT, "rgb_light_diffuse[i]", "vec3", "l_diffuse[i]*light_sources[i].diffuse.rgb", 0 },
	{ FRAGMENT, "l_diffuse[i]", "float", "max(dot(n_zzz_normal, n_zzz_light_dir[i]), 0.0)", 0 },

	{ FRAGMENT, "l_shadow", "float", "mix(1.0, shadow_sample, shadow_darkness)", 0 },
	{ FRAGMENT, "shadow_sample", "float", "shadow2D(shadow, shd_vertex)", 0 },

	{ FRAGMENT, "zzz_reflect_dir", "vec3", "reflect(zzz_incident_dir, n_zzz_normal)", 0 },
	{ FRAGMENT, "n_zzz_half_vec[i]", "vec3", "normalize(zzz_light_dir[i]-zzz_incident_dir)", 0 },
	{ FRAGMENT, "n_zzz_light_dir[i]", "vec3", "normalize(zzz_light_dir[i])", 0 },
	{ FRAGMENT, "n_tbn_normal", "vec3", "normal_sample*2.0-1.0", "n" },
	{ FRAGMENT, "n_eye_normal", "vec3", "normalize(eye_normal)", "!n" },
	{ FRAGMENT, "normal_sample", "vec3", "texture2D(normal_map, texture_coord).xyz", 0 },
	{ FRAGMENT, "diffuse_sample", "vec4", "texture2D(diffuse_map, texture_coord)", 0 },

	{ VERTEX, "gl_Position", "vec4", "projection_matrix*eye_vertex", 0 },
	{ VERTEX, "gl_ClipDistance[i]", "float", "dot(eye_vertex, clip_planes[i].equation)", "c" },
	{ VERTEX, "fog_coord", "float", "-eye_vertex.z", 0 },
	{ VERTEX, "shd_vertex", "vec3", "(shd_eye_matrix*eye_vertex).xyz", 0 },
	{ VERTEX, "tbn_sky_dir", "vec3", "eye_sky_dir*eye_tbn_matrix", "n" },
	{ VERTEX, "tbn_light_dir[i]", "vec3", "eye_light_dir[i]*eye_tbn_matrix", 0 },
	{ VERTEX, "eye_light_dir[i]", "vec3", "normalize(eye_light_position[i].xyz-eye_vertex.xyz*eye_light_position[i].w)", 0 },
	{ VERTEX, "eye_light_position[i]", "vec4", "light_sources[i].position", 0 },
	{ VERTEX, "tbn_incident_dir", "vec3", "eye_incident_dir*eye_tbn_matrix", 0 },
	{ VERTEX, "eye_incident_dir", "vec3", "normalize(eye_vertex.xyz)", 0 },
	{ VERTEX, "eye_tbn_matrix", "mat3", "mat3(eye_tangent, eye_binormal, eye_normal)", 0 },
	{ VERTEX, "eye_vertex", "vec4", "eye_obj_matrix*vertex", 0 },
	{ VERTEX, "eye_normal", "vec3", "eye_obj_normal_matrix*normal", 0 },
	{ VERTEX, "eye_tangent", "vec3", "eye_obj_normal_matrix*tangent", 0 },
	{ VERTEX, "eye_binormal", "vec3", "eye_obj_normal_matrix*binormal", 0 },
	{ VERTEX, "texture_coord", "vec2", "texcoord.xy", 0 },

	{ ATTRIBUTE, "vertex", "vec4", "gl_Vertex", 0 },
	{ ATTRIBUTE, "texcoord", "vec4", "gl_MultiTexCoord0", 0 },
	{ ATTRIBUTE, "color", "vec4", "gl_Color", 0 },
	{ ATTRIBUTE, "normal", "vec3", "gl_Normal", 0 },
	{ ATTRIBUTE, "tangent", "vec3", 0, 0 },
	{ ATTRIBUTE, "binormal", "vec3", 0, 0 },

	{ UNIFORM, "diffuse_map", "sampler2D", 0, 0 },
	{ UNIFORM, "shadow", "sampler2DShadow", 0, 0 },
	{ UNIFORM, "ShadowMap::shadow_darkness", "float", 0, 0 },
	{ UNIFORM, "normal_map", "sampler2D", 0, 0 },
	{ UNIFORM, "environment", "samplerCube", 0, 0 },
	{ UNIFORM, "EnvMap::env_eye_matrix", "mat3", 0, 0 },
	{ UNIFORM, "Material::reflectivity", "float", 0, 0 },
	{ UNIFORM, "eye_obj_matrix", "mat4", "gl_ModelViewMatrix", 0 },
	{ UNIFORM, "eye_obj_normal_matrix", "mat3", "gl_NormalMatrix", 0 },
	{ UNIFORM, "Transform::projection_matrix", "mat4", "gl_ProjectionMatrix", 0 },
	{ UNIFORM, "ShadowMap::shd_eye_matrix", "mat4", 0, 0 },
	{ UNIFORM, "Lighting::light_sources", "LightSourceParameters[MAX_LIGHTS]", "gl_LightSource[i]", 0 },
	{ UNIFORM, "Lighting::ambient_color", "vec4", "gl_LightModel.ambient", 0 },
	{ UNIFORM, "Lighting::sky_color", "vec4", 0, 0 },
	{ UNIFORM, "Lighting::eye_sky_dir", "vec3", 0, 0 },
	{ UNIFORM, "Lighting::horizon_limit", "float", 0, 0 },
	{ UNIFORM, "Lighting::fog_color", "vec4", "gl_Fog.color", 0 },
	{ UNIFORM, "Lighting::fog_density", "float", "gl_Fog.density", 0 },
	{ UNIFORM, "Material::material", "MaterialParameters", "gl_FrontMaterial", 0 },
	{ UNIFORM, "Clipping::clip_planes", "ClipPlane[MAX_CLIP_PLANES]", 0, 0 },

	{ TYPE, "LightSourceParameters", "struct { vec4 position; vec4 diffuse; vec4 specular; }", "gl_LightSourceParameters", 0 },
	{ TYPE, "MaterialParameters", "struct { vec4 ambient; vec4 diffuse; vec4 specular; float shininess; }", "gl_MaterialParameters", 0 },
	{ TYPE, "ClipPlane", "struct { vec4 equation; }", 0, 0 },

	// Terminator entry
	{ NO_SCOPE, 0, 0, 0, 0 }
};

const char ProgramBuilder::interfaces[] = { 0, 0, 0, 0, 0, 'v', 'g', 0 };

ProgramBuilder::ProgramBuilder(const StandardFeatures &f):
	features(f),
	feature_flags(features.create_flags()),
	enabled_scopes((1<<N_SCOPES)-1),
	optimize(true)
{
	if(!features.custom.empty())
	{
		const char *whitespace = " \t\n";
		string::size_type start = 0;
		while(1)
		{
			start = features.custom.find_first_not_of(whitespace, start);
			if(start==string::npos)
				break;

			string::size_type semicolon = features.custom.find(';', start);
			if(semicolon==start)
			{
				++start;
				continue;
			}
			else if(semicolon==string::npos)
				throw invalid_variable_definition(features.custom.substr(start));

			string::size_type equals = features.custom.find('=', start);
			if(equals>semicolon)
				equals = string::npos;

			VariableDefinition var;
			string::size_type decl_end = min(equals, semicolon);
			for(unsigned i=0;; ++i)
			{
				string::size_type word_end = features.custom.find_first_of(whitespace, start);
				word_end = min(word_end, decl_end);
				features.custom[word_end] = 0;

				const char *word = &features.custom[start];
				if(i==0)
				{
					if(!strcmp(word, "uniform"))
						var.scope = UNIFORM;
					else if(!strcmp(word, "attribute"))
						var.scope = ATTRIBUTE;
					else if(!strcmp(word, "vertex"))
						var.scope = VERTEX;
					else if(!strcmp(word, "fragment"))
						var.scope = FRAGMENT;
					else if(!strcmp(word, "function"))
						var.scope = FUNCTION;
					else
						throw invalid_variable_definition(word);
				}
				else if(i==1)
					var.type = word;
				else if(i==2)
					var.name = word;

				start = features.custom.find_first_not_of(whitespace, word_end+1);
				if(start>=decl_end || (var.scope==FUNCTION && features.custom[start]=='('))
					break;
			}

			if(equals!=string::npos)
			{
				if(var.scope==FUNCTION)
					throw invalid_variable_definition("function with expression");
				start = features.custom.find_first_not_of(whitespace, equals+1);
				if(start>=semicolon)
					throw invalid_variable_definition("no expression");
				features.custom[semicolon] = 0;
				var.expression = &features.custom[start];
			}
			else if(var.scope==FUNCTION)
			{
				string::size_type left_paren = features.custom.find('(', start);
				string::size_type right_paren = features.custom.find(')', start);
				if(left_paren>semicolon || right_paren>semicolon)
					throw invalid_variable_definition("no argument list");
				features.custom[right_paren] = 0;
				var.expression = &features.custom[left_paren+1];
			}
			else
				var.expression = 0;

			var.flags = 0;
			for(const VariableDefinition *j=standard_variables; j->name; ++j)
				if(!strcmp(var.name, j->name))
					var.flags = "o";

			custom_variables.push_front(var);

			start = semicolon+1;
		}
	}

	if(!features.geometry)
		enabled_scopes &= ~(1<<GEOMETRY);

	if(!features.legacy)
	{
		aliases["texture1D"] = "texture";
		aliases["texture2D"] = "texture";
		aliases["texture3D"] = "texture";
		aliases["textureCube"] = "texture";
		aliases["shadow2D"] = "texture";
	}
	else
		aliases["shadow2D"] = "shadow2D(...).r";
}

void ProgramBuilder::set_optimize(bool o)
{
	optimize = o;
}

Program *ProgramBuilder::create_program() const
{
	Program *prog = new Program;
	add_shaders(*prog);
	return prog;
}

void ProgramBuilder::add_shaders(Program &prog) const
{
	list<ShaderVariable> variables;
	list<ShaderVariable *> resolved_vars;

	if(features.clipping)
		variables.push_back(ShaderVariable("gl_ClipDistance[i]"));
	variables.push_front(ShaderVariable("gl_Position"));
	variables.push_front(ShaderVariable(features.legacy ? "gl_FragColor" : "frag_color"));

	list<VariableDefinition>::const_iterator next_custom = custom_variables.begin();
	for(const VariableDefinition *i=standard_variables; i->name; )
	{
		const VariableDefinition *def = 0;
		if(next_custom!=custom_variables.end() && (!strcmp(next_custom->name, i->name) || !next_custom->flags))
		{
			def = &*next_custom;
			++next_custom;
		}
		else
		{
			def = i;
			++i;

			// Skip over anything that isn't used with the supplied flags
			if(def->flags && !evaluate_flags(def->flags))
				continue;
		}

		if(def->scope==TYPE)
		{
			for(list<ShaderVariable *>::iterator j=resolved_vars.begin(); j!=resolved_vars.end(); ++j)
				if(!(*j)->type && name_match(def->name, (*j)->variable->type))
					(*j)->resolve_type(*def);

			continue;
		}

		const char *def_uq_name = unqualified_name(def->name);

		// See if this variable can satisfy any unresolved variables
		ShaderVariable *last_resolved = 0;
		for(list<ShaderVariable>::iterator j=variables.begin(); j!=variables.end(); ++j)
		{
			if(j->variable)
				continue;

			if(!name_match(def_uq_name, j->resolved_name.c_str()))
				continue;

			if(last_resolved)
			{
				/* We've already resolved a non-fuzzy variable in this iteration.
				If there are multiple variables that can be resolved, they refer
				to the same variable. */
				j->resolve(*last_resolved);
				continue;
			}

			j->resolve(*def);
			resolved_vars.push_front(&*j);
			if(!j->fuzzy_space)
				last_resolved = &*j;

			if(!def->expression)
				continue;

			vector<string> identifiers = extract_identifiers(def->expression);
			for(vector<string>::const_iterator k=identifiers.begin(); k!=identifiers.end(); ++k)
			{
				// Use an existing variable if possible, but only if it's not fuzzy
				ShaderVariable *var = 0;
				for(list<ShaderVariable>::iterator l=variables.begin(); (!var && l!=variables.end()); ++l)
					if(!l->fuzzy_space && l->resolved_name==*k)
						var = &*l;

				if(!var)
				{
					variables.push_back(ShaderVariable(*k));
					var = &variables.back();
				}
				j->add_reference(*var);
			}
		}
	}

	// Array sizes need to be resolved for inline processing
	for(list<ShaderVariable>::iterator i=variables.end(); i!=variables.begin(); )
		(--i)->resolve_array(features);

	bool legacy_variables = evaluate_flags("g");
	for(list<ShaderVariable *>::const_iterator i=resolved_vars.begin(); i!=resolved_vars.end(); ++i)
		(*i)->check_inline(legacy_variables, !optimize);

	prog.attach_shader_owned(new VertexShader(create_source(resolved_vars, VERTEX)));
	prog.attach_shader_owned(new FragmentShader(create_source(resolved_vars, FRAGMENT)));

	if(!features.legacy)
	{
		// OpenGL ES does not support binding fragment shader outputs
		if(get_gl_api()!=OPENGL_ES2)
			prog.bind_fragment_data(0, "frag_color");

		prog.bind_attribute(VERTEX4, "vertex");
		if(features.lighting)
			prog.bind_attribute(NORMAL3, "normal");
		else if(features.material)
			prog.bind_attribute(COLOR4_FLOAT, "color");
		if(features.texture || features.normal_map)
			prog.bind_attribute(TEXCOORD4, "texcoord");
	}
	if(features.normal_map)
	{
		prog.bind_attribute(get_component_type(TANGENT3), "tangent");
		prog.bind_attribute(get_component_type(BINORMAL3), "binormal");
	}
}

string ProgramBuilder::create_source(const list<ShaderVariable *> &variables, VariableScope scope) const
{
	string source;

	bool use_blocks = !features.legacy && ARB_uniform_buffer_object;

	if(!features.legacy)
	{
		if(get_gl_api()==OPENGL_ES2)
		{
			if(use_blocks)
				source += "#version 300 es\n";
		}
		else
		{
			source += "#version 130\n";
			if(use_blocks)
				source += "#extension GL_ARB_uniform_buffer_object: require\n";
		}
	}

	if(get_gl_api()==OPENGL_ES2)
		source += "precision mediump float;\n";

	set<const VariableDefinition *> declared_types;
	set<string> uniform_blocks;
	for(list<ShaderVariable *>::const_iterator i=variables.begin(); i!=variables.end(); ++i)
	{
		if(!(*i)->is_referenced_from(scope))
			continue;

		if((*i)->variable->scope==FUNCTION)
			source += format("%s;\n", (*i)->create_declaration());
		else if((*i)->variable->scope==UNIFORM && !(*i)->inlined)
		{
			if((*i)->type && !declared_types.count((*i)->type))
			{
				source += format("%s;\n", (*i)->create_type_declaration());
				declared_types.insert((*i)->type);
			}

			if(!(*i)->resolved_block.empty() && use_blocks)
				uniform_blocks.insert((*i)->resolved_block);
			else
				source += format("uniform %s;\n", (*i)->create_declaration());
		}
	}

	for(set<string>::const_iterator i=uniform_blocks.begin(); i!=uniform_blocks.end(); ++i)
	{
		source += format("uniform %s\n{\n", *i);
		for(list<ShaderVariable *>::const_iterator j=variables.begin(); j!=variables.end(); ++j)
			if((*j)->resolved_block==*i)
				source += format("\t%s;\n", (*j)->create_declaration());
		source += "};\n";
	}

	/* Interface variables need to have global declarations. */
	for(list<ShaderVariable *>::const_iterator i=variables.begin(); i!=variables.end(); ++i)
	{
		if(!(*i)->resolved_name.compare(0, 3, "gl_"))
			continue;

		InterfaceFlags iface = (*i)->get_interface_flags(scope);

		if(iface&INPUT)
		{
			const char *qualifier = (features.legacy ? scope==VERTEX ? "attribute" : "varying" : "in");
			source += format("%s %s;\n", qualifier, (*i)->create_declaration(interfaces[previous_scope(scope, enabled_scopes)]));
		}

		if(iface&OUTPUT)
		{
			const char *qualifier = (features.legacy ? "varying" : "out");
			source += format("%s %s;\n", qualifier, (*i)->create_declaration(interfaces[scope]));
		}
	}

	// The clip distance array must be declared manually to give it a size
	if(scope==VERTEX && features.clipping)
		source += format("out float gl_ClipDistance[%d];\n", features.max_clip_planes);

	source += "void main()\n{\n";

	list<ShaderVariable *> loop_vars;
	unsigned loop_size = 0;
	for(list<ShaderVariable *>::const_iterator i=variables.begin(); i!=variables.end(); ++i)
	{
		if(!loop_vars.empty() && !loop_vars.back()->in_loop && (*i)->array_size!=loop_size)
		{
			/* Declare all variables that need to be visible outside the loop.
			Output variables are already declared. */
			for(list<ShaderVariable *>::const_iterator j=loop_vars.begin(); j!=loop_vars.end(); ++j)
			{
				InterfaceFlags iface = (*j)->get_interface_flags(scope);
				if(!(*j)->in_loop && !(iface&OUTPUT))
					source += format("\t%s;\n", (*j)->create_declaration());
			}

			source += format("\tfor(int i=0; i<%d; ++i)\n\t{\n", loop_size);
			for(list<ShaderVariable *>::const_iterator j=loop_vars.begin(); j!=loop_vars.end(); ++j)
			{
				if((*j)->variable->scope==scope && !(*j)->inlined)
				{
					string decl;
					if((*j)->in_loop)
						decl = (*j)->create_declaration(0, true);
					else
					{
						decl = (*j)->resolved_name;
						if(!(*j)->array_sum)
							decl += "[i]";
					}
					const char *oper = ((*j)->array_sum ? "+=" : "=");
					source += format("\t\t%s %s %s;\n", decl, oper, create_expression(**j, "i"));
				}

				InterfaceFlags iface = (*j)->get_interface_flags(scope);
				if((iface&(OUTPUT|GOAL))==OUTPUT)
				{
					string expr = ((*j)->inlined ? create_expression(**j, "i") : (*j)->resolved_name+"[i]");
					source += format("\t\t%c_%s[i] = %s;\n", interfaces[scope], (*j)->resolved_name, expr);
				}
			}
			source += "\t}\n";

			loop_vars.clear();
		}

		InterfaceFlags iface = (*i)->get_interface_flags(scope);

		if((*i)->array_size>1 || ((*i)->array_size==1 && (iface&GOAL)))
		{
			if((*i)->variable->scope==scope || (iface&OUTPUT))
			{
				loop_size = (*i)->array_size;
				loop_vars.push_back(*i);
			}
			continue;
		}

		if((*i)->variable->scope==scope && !(*i)->inlined)
		{
			string decl = ((iface&GOAL) ? (*i)->resolved_name : (*i)->create_declaration());
			source += format("\t%s = %s;\n", decl, create_expression(**i));
		}

		if((iface&(OUTPUT|GOAL))==OUTPUT)
		{
			string expr = ((*i)->inlined ? create_expression(**i) : (*i)->resolved_name);
			source += format("\t%c_%s = %s;\n", interfaces[scope], (*i)->resolved_name, expr);
		}
	}

	source += '}';

	return source;
}

bool ProgramBuilder::evaluate_flags(const char *flags) const
{
	if(!flags)
		return true;

	bool cond = true;
	char oper = '&';
	for(const char *i=flags; *i; ++i)
	{
		if(*i>='a' && *i<='z')
		{
			bool found = (feature_flags.find(*i)!=string::npos);
			if(oper=='|')
				cond = (cond || found);
			else if(oper=='!')
				cond = (cond && !found);
			else if(oper=='&')
				cond = (cond && found);
			oper = '&';
		}
		else
			oper = *i;
	}

	return cond;
}

const char *ProgramBuilder::unqualified_name(const char *name)
{
	for(const char *p=name; *p; ++p)
		if(*p==':' && *++p==':')
			name = p+1;
	return name;
}

ProgramBuilder::MatchType ProgramBuilder::name_match(const char *n1, const char *n2, const char **space)
{
	int i = 0;
	int zzz = -1;
	int zside = 0;
	while(*n1 && *n2)
	{
		if(*n1==*n2 || *n1=='z' || *n2=='z')
		{
			if(*n1!=*n2)
			{
				int side = (*n1=='z' ? 1 : 2);
				if(zzz<0)
				{
					zzz = i;
					zside = side;
					if(space)
					{
						if(*n1=='z')
							*space = n2;
						else
							*space = n1;
					}
				}
				else if(i>=zzz+3 || side!=zside)
					return NO_MATCH;
			}
		}
		else
			return NO_MATCH;
		++n1;
		++n2;
		++i;
	}
	return (!*n1 && !*n2) ? (zzz>=0 ? FUZZY : EXACT) : ((*n1=='[' || *n2=='[') ? ARRAY : NO_MATCH);
}

bool ProgramBuilder::parse_identifier(const char *ptr, unsigned &start, unsigned &length)
{
	bool found = false;
	bool member = false;
	bool subscript = false;
	for(const char *i=ptr;; ++i)
	{
		if(!found)
		{
			if(!*i)
				return false;
			if(isalpha(*i) || *i=='_')
			{
				if(!member)
				{
					start = i-ptr;
					found = true;
				}
			}
			else if(*i=='.')
				member = true;
			else
				member = false;
		}
		else
		{
			if(subscript)
			{
				if(*i==']')
				{
					length = i+1-(ptr+start);
					return true;
				}
				else if(!isalpha(*i) || i>ptr+start+length+1)
					return true;
			}
			else if(!isalnum(*i) && *i!='_')
			{
				length = i-(ptr+start);
				if(*i=='[')
					subscript = true;
				else
					return true;
			}
		}
	}
}

vector<string> ProgramBuilder::extract_identifiers(const char *expression)
{
	vector<string> result;
	const char *ptr = expression;
	unsigned start = 0;
	unsigned length = 0;
	while(parse_identifier(ptr, start, length))
	{
		result.push_back(string(ptr+start, length));
		ptr += start+length;
	}
	return result;
}

string ProgramBuilder::replace_identifiers(const char *expression, const map<string, string> &replace_map, bool with_functions)
{
	string result;
	const char *ptr = expression;
	unsigned start = 0;
	unsigned length = 0;
	while(parse_identifier(ptr, start, length))
	{
		result.append(ptr, start);
		string identifier(ptr+start, length);
		map<string, string>::const_iterator i = replace_map.find(identifier);
		if(i!=replace_map.end())
		{
			if(with_functions && ptr[start+length]=='(')
			{
				string::size_type lparen = i->second.find('(');
				string::size_type rparen = i->second.rfind(')');
				if(lparen!=string::npos && rparen!=string::npos)
				{
					unsigned level = 1;
					unsigned j;
					for(j=start+length+1; (ptr[j] && level); ++j)
					{
						level += (ptr[j]=='(');
						level -= (ptr[j]==')');
					}

					if(!level)
					{
						string subexpr(ptr+start+length, ptr+j);
						result += i->second.substr(0, lparen);
						result += replace_identifiers(subexpr.c_str(), replace_map, with_functions);
						result += i->second.substr(rparen+1);
						ptr += j;
						continue;
					}
				}
			}

			result += i->second;
		}
		else
			result += identifier;
		ptr += start+length;
	}
	result += ptr;
	return result;
}

ProgramBuilder::VariableScope ProgramBuilder::previous_scope(VariableScope scope, unsigned enabled_scopes)
{
	while(scope!=NO_SCOPE)
	{
		scope = static_cast<VariableScope>(scope-1);
		if(enabled_scopes&(1<<scope))
			break;
	}

	return scope;
}

string ProgramBuilder::create_expression(const ShaderVariable &var, const char *loop) const
{
	string expr = var.create_expression(enabled_scopes, loop);
	return replace_identifiers(expr.c_str(), aliases, true);
}


ProgramBuilder::StandardFeatures::StandardFeatures():
	texture(false),
	material(false),
	lighting(false),
	max_lights(1),
	skylight(false),
	fog(false),
	specular(false),
	normal_map(false),
	shadow(false),
	reflection(false),
	clipping(false),
	max_clip_planes(1),
	geometry(false)
{
	if(get_gl_api()==OPENGL_ES2)
		legacy = get_glsl_version()<Version(3, 0);
	else
		legacy = get_glsl_version()<Version(1, 30);
}

string ProgramBuilder::StandardFeatures::create_flags() const
{
	string flags;
	if(texture)
		flags += 't';
	if(material)
		flags += 'm';
	if(lighting)
	{
		flags += 'l';
		if(skylight)
			flags += 'y';
		if(specular)
			flags += 'p';
		if(normal_map)
			flags += 'n';
	}
	if(fog)
		flags += 'f';
	if(shadow)
		flags += 's';
	if(reflection)
		flags += 'e';
	if(clipping)
		flags += 'c';
	if(legacy && get_gl_api()==OPENGL)
		flags += 'g';

	return flags;
}


ProgramBuilder::ShaderVariable::ShaderVariable(const std::string &n):
	name(n),
	variable(0),
	type(0),
	resolved_name(n),
	fuzzy_space(name.find("zzz")!=string::npos),
	array_sum(false),
	array_size(0),
	inlined(false),
	inline_parens(false),
	in_loop(false)
{
	string::size_type bracket = name.find('[');
	if(bracket!=string::npos)
		array_subscript = name.substr(bracket+1, name.size()-bracket-2);
}

void ProgramBuilder::ShaderVariable::resolve(const VariableDefinition &var)
{
	variable = &var;
	const char *space = 0;
	const char *var_uq_name = unqualified_name(variable->name);
	MatchType match = name_match(var_uq_name, resolved_name.c_str(), &space);

	if(var_uq_name!=variable->name)
		resolved_block.assign(variable->name, var_uq_name-2);

	if(match==FUZZY)
		resolve_space(string(space, 3));
	else if(match==ARRAY)
	{
		if(array_subscript.empty())
			array_sum = true;
		else if(var.scope==UNIFORM)
		{
			const char *bracket = strrchr(variable->type, '[');
			if(bracket)
				array_subscript = string(bracket+1, strlen(bracket)-2);
		}
	}
}

void ProgramBuilder::ShaderVariable::resolve(ShaderVariable &var)
{
	for(list<ShaderVariable *>::iterator i=referenced_by.begin(); i!=referenced_by.end(); ++i)
		(*i)->update_reference(*this, var);
	var.referenced_by.insert(var.referenced_by.end(), referenced_by.begin(), referenced_by.end());
}

void ProgramBuilder::ShaderVariable::resolve_type(const VariableDefinition &var)
{
	type = &var;
}

void ProgramBuilder::ShaderVariable::resolve_space(const string &space)
{
	if(fuzzy_space)
	{
		resolved_space = space;

		string::size_type zzz = resolved_name.find("zzz");
		resolved_name.replace(zzz, 3, resolved_space);
		fuzzy_space = false;

		// Resolving the space could have affected other variables that use this one
		for(list<ShaderVariable *>::iterator i=referenced_by.begin(); i!=referenced_by.end(); ++i)
			(*i)->resolve_space(space);
	}

	for(list<ShaderVariable *>::iterator i=referenced_vars.begin(); i!=referenced_vars.end(); ++i)
		if((*i)->fuzzy_space)
			(*i)->resolve_space(space);
}

void ProgramBuilder::ShaderVariable::resolve_array(const StandardFeatures &feat, unsigned size_hint)
{
	if(array_size)
		return;
	if(!array_sum && array_subscript.empty())
		return;

	if(!array_subscript.empty())
	{
		string::size_type bracket = resolved_name.find('[');
		if(bracket!=string::npos)
			resolved_name = resolved_name.substr(0, bracket);
	}

	if(variable && variable->scope==UNIFORM)
	{
		if(array_subscript=="MAX_LIGHTS")
			array_size = feat.max_lights;
		else if(array_subscript=="MAX_CLIP_PLANES")
			array_size = feat.max_clip_planes;
		else if(isnumrc(array_subscript))
			array_size = lexical_cast<unsigned>(array_subscript);
		else
			throw invalid_variable_definition("invalid array size");
	}

	if(!array_size)
	{
		for(list<ShaderVariable *>::const_iterator i=referenced_vars.begin(); i!=referenced_vars.end(); ++i)
			if((*i)->array_size)
			{
				array_size = (*i)->array_size;
				break;
			}
	}

	if(!array_size && size_hint)
		array_size = size_hint;

	if(array_size)
	{
		for(list<ShaderVariable *>::const_iterator i=referenced_vars.begin(); i!=referenced_vars.end(); ++i)
			if(!(*i)->array_subscript.empty() && !(*i)->array_size)
				(*i)->resolve_array(feat, array_size);
	}
}

void ProgramBuilder::ShaderVariable::add_reference(ShaderVariable &var)
{
	referenced_vars.push_back(&var);
	var.referenced_by.push_back(this);
	if(var.fuzzy_space && !resolved_space.empty())
		var.resolve_space(resolved_space);
}

void ProgramBuilder::ShaderVariable::update_reference(ShaderVariable &from, ShaderVariable &to)
{
	replace(referenced_vars.begin(), referenced_vars.end(), &from, &to);
	replace(referenced_by.begin(), referenced_by.end(), &from, &to);
	if(from.fuzzy_space && !to.fuzzy_space && !to.resolved_space.empty())
		resolve_space(to.resolved_space);
}

void ProgramBuilder::ShaderVariable::check_inline(bool allow_legacy, bool trivial_only)
{
	if(variable->expression && variable->scope>=UNIFORM)
	{
		if(array_sum && array_size>1)
			return;
		if(!allow_legacy && !strncmp(variable->expression, "gl_", 3))
			return;

		// Never inline goal variables
		if(referenced_by.empty())
			return;

		// Inline an expression consisting of a single identifier
		unsigned start, length;
		if(parse_identifier(variable->expression, start, length))
			if(start==0 && variable->expression[length]==0)
			{
				inlined = true;
				return;
			}

		if(trivial_only)
			return;

		/* If all references to the variable come from arrays in the same scope
		and of the same size, the variable can be embedded in the loop. */
		in_loop = (array_size>1 && !array_sum);
		for(list<ShaderVariable *>::const_iterator i=referenced_by.begin(); i!=referenced_by.end(); ++i)
			if((*i)->variable->scope!=variable->scope || (*i)->array_size!=array_size)
				in_loop = false;
		
		/* Count all refs to this variable.  Refs from array variables count once
		per loop iteration. */
		unsigned total_refs = 0;
		unsigned in_scope_refs = 0;
		for(list<ShaderVariable *>::const_iterator i=referenced_by.begin(); i!=referenced_by.end(); ++i)
		{
			unsigned count = max((*i)->array_size*!in_loop, 1U);
			total_refs += count;
			if((*i)->variable->scope==variable->scope)
				in_scope_refs += count;
		}

		/* Inline if there's only one ref, or if all refs are in other scopes.
		In the latter case, the actual inlining will happen in the interface
		variable assignment. */
		if(total_refs==1 || in_scope_refs==0)
		{
			inlined = true;
			unsigned level = 0;
			for(const char *c=variable->expression; (!inline_parens && *c); ++c)
			{
				if(*c=='(')
					++level;
				else if(*c==')')
					--level;
				else if(level==0 && !isalnum(*c) && *c!='_' && *c!='.')
					inline_parens = true;
			}
		}
	}
}

bool ProgramBuilder::ShaderVariable::is_referenced_from(VariableScope scope) const
{
	for(list<ShaderVariable *>::const_iterator i=referenced_by.begin(); i!=referenced_by.end(); ++i)
		if((*i)->variable->scope==scope)
			return true;
	return false;
}

ProgramBuilder::InterfaceFlags ProgramBuilder::ShaderVariable::get_interface_flags(VariableScope scope) const
{
	/* Uniforms are available to all stages and are not passed through
	interfaces */
	if(variable->scope<=UNIFORM)
		return NO_INTERFACE;

	int flags = NO_INTERFACE;

	for(list<ShaderVariable *>::const_iterator i=referenced_by.begin(); i!=referenced_by.end(); ++i)
	{
		/* Variables used in a later scope than they are declared in need to go
		through the interface */
		if((*i)->variable->scope>scope && variable->scope<=scope)
			flags |= OUTPUT;
		if((*i)->variable->scope>=scope && variable->scope<scope)
			if(!inlined || variable->scope!=ATTRIBUTE || scope!=VERTEX)
				flags |= INPUT;
	}

	// Variables without any references are goals and also outputs.
	if(referenced_by.empty() && variable->scope==scope)
		flags |= OUTPUT|GOAL;

	return static_cast<InterfaceFlags>(flags);
}

string ProgramBuilder::ShaderVariable::create_type_declaration() const
{
	if(!type)
		throw logic_error("no type");

	if(!strncmp(type->type, "struct", 6))
	{
		const char *brace = strchr(type->type, '{');
		if(brace)
			return format("struct %s %s", type->name, brace);
	}

	throw invalid_variable_definition("invalid typedef");
}

string ProgramBuilder::ShaderVariable::create_declaration(char iface, bool loop) const
{
	if(variable->scope==FUNCTION)
		return format("%s %s(%s)", variable->type, resolved_name, variable->expression);

	if(variable->scope==UNIFORM && !array_subscript.empty())
	{
		const char *bracket = strrchr(variable->type, '[');
		if(bracket)
			return format("%s %s[%d]", string(variable->type, bracket), resolved_name, array_size);
	}

	string array;
	if(!array_sum && array_size>1 && !loop)
		array = format("[%d]", array_size);

	if(iface)
		return format("%s %c_%s%s", variable->type, iface, resolved_name, array);
	else
		return format("%s %s%s", variable->type, resolved_name, array);
}

string ProgramBuilder::ShaderVariable::create_replacement(VariableScope from_scope, unsigned scopes, const char *loop) const
{
	string replacement = resolved_name;
	InterfaceFlags iface = NO_INTERFACE;
	if(variable)
	{
		iface = get_interface_flags(from_scope);
		VariableScope prev_scope = previous_scope(from_scope, scopes);
		if((iface&INPUT) && interfaces[prev_scope])
			replacement = format("%c_%s", interfaces[prev_scope], replacement);
		else if(inlined)
		{
			replacement = create_expression(scopes, loop);
			if(inline_parens)
				replacement = "("+replacement+")";
			return replacement;
		}
	}

	// Add an array subscript, unless the variable is embedded in a loop
	if(!array_subscript.empty() && !in_loop)
	{
		if(loop)
			return format("%s[%s]", replacement, loop);
		else if(!variable || variable->scope==UNIFORM)
			return replacement+"[0]";
	}

	return replacement;
}

string ProgramBuilder::ShaderVariable::create_expression(unsigned scopes, const char *loop) const
{
	map<string, string> replace_map;
	for(list<ShaderVariable *>::const_iterator i=referenced_vars.begin(); i!=referenced_vars.end(); ++i)
	{
		string replacement = (*i)->create_replacement(variable->scope, scopes, loop);
		if(replacement!=(*i)->name)
			replace_map[(*i)->name] = replacement;
	}

	if(replace_map.empty())
		return variable->expression;
	else
		return replace_identifiers(variable->expression, replace_map);
}


ProgramBuilder::StandardFeatures::Loader::Loader(StandardFeatures &f):
	DataFile::ObjectLoader<StandardFeatures>(f)
{
	add("clipping",  &StandardFeatures::clipping);
	add("custom",    &StandardFeatures::custom);
	add("fog",       &StandardFeatures::fog);
	add("geometry",  &StandardFeatures::geometry);
	add("lighting",  &StandardFeatures::lighting);
	add("material",  &StandardFeatures::material);
	add("max_clip_planes", &StandardFeatures::max_clip_planes);
	add("max_lights", &StandardFeatures::max_lights);
	add("normal_map", &StandardFeatures::normal_map);
	add("reflection", &StandardFeatures::reflection);
	add("shadow",    &StandardFeatures::shadow);
	add("skylight",  &StandardFeatures::skylight);
	add("specular",  &StandardFeatures::specular);
	add("texture",   &StandardFeatures::texture);

	// Deprecated
	add("normalmap", &StandardFeatures::normal_map);
}

} // namespace GL
} // namespace Msp
